package com.rp.util.sort;

import com.rp.util.sort.ISort;
import com.rp.util.sort.SortException;
import com.rp.util.ReflectionUtil;
import java.util.Comparator;
import java.util.List;

public final class HeapSort implements ISort
{
	
	public HeapSort ()
	{}
	
	public void sort (int[] arr)
	{
		sort (arr, 0, arr.length-1);
	}	
	
	public void sort (int[] arr, int begin, int end)
	{
		buildHeap (arr, begin, end);
		int temp;
		for (int i = (end-begin); i > begin; i--)
		{
			temp = arr[begin];
			arr[begin] = arr[i];
			arr[i] = temp;
			heapify (arr, begin, i-1);
		}	
	}
	
	protected void buildHeap (int[] arr, int begin, int end)
	{
		int n = (end-begin);
		for (int i = (n / 2); i >= begin; i--)
		{
			heapify (arr, i, n);	
		}
	}
	
	protected void heapify (int[] arr, int i, int heapSize)
	{
		int left = 2*i;
		int right = (2*i)+1;
		int largest;
		
		if (left <= heapSize && arr[left] > arr[i])
		{
			largest = left;
		}
		else
		{
			largest = i;	
		}
		if (right <= heapSize && arr[right] > arr[largest])
		{
			largest = right;	
		} 
		if (largest != i)
		{
			int temp = arr[i];
			arr[i] = arr[largest];
			arr[largest] = temp;
			heapify (arr, largest, heapSize);	
		}
	}

	public void sort (String[] arr)
	{
		sort (arr, 0, arr.length-1);	
	}
	
	public void sort (String[] arr, int begin, int end)
	{
		buildHeap (arr, begin, end);
		String temp;
		for (int i = (end-begin); i > begin; i--)
		{
			temp = arr[begin];
			arr[begin] = arr[i];
			arr[i] = temp;
			heapify (arr, begin, i-1);
		}	
	}
	
	protected void buildHeap (String[] arr, int begin, int end)
	{
		int n = (end-begin);
		for (int i = (n / 2); i >= begin; i--)
		{
			heapify (arr, i, n);	
		}
	}
	
	protected void heapify (String[] arr, int i, int heapSize)
	{
		int left = 2*i;
		int right = (2*i)+1;
		int largest;
		
		if (left <= heapSize && arr[left].compareToIgnoreCase(arr[i]) > 0)
		{
			largest = left;
		}
		else
		{
			largest = i;	
		}
		if (right <= heapSize && arr[right].compareToIgnoreCase (arr[largest]) > 0)
		{
			largest = right;	
		} 
		if (largest != i)
		{
			String temp = arr[i];
			arr[i] = arr[largest];
			arr[largest] = temp;	
			heapify (arr, largest, heapSize);	
		}
	}

	public void sort (long[] arr)
	{
		sort (arr, 0, arr.length-1);
	}
	
	public void sort (long[] arr, int begin, int end)
	{
		buildHeap (arr, begin, end);
		long temp;
		for (int i = (end-begin); i > begin; i--)
		{
			temp = arr[begin];
			arr[begin] = arr[i];
			arr[i] = temp;
			heapify (arr, begin, i-1);
		}	
	}
	
	protected void buildHeap (long[] arr, int begin, int end)
	{
		int n = end - begin;
		for (int i = (n / 2); i >= begin; i--)
		{
			heapify (arr, i, n);	
		}
	}
	
	protected void heapify (long[] arr, int i, int heapSize)
	{
		int left = 2*i;
		int right = (2*i)+1;
		int largest;
		
		if (left <= heapSize && arr[left] > arr[i])
		{
			largest = left;
		}
		else
		{
			largest = i;	
		}
		if (right <= heapSize && arr[right] > arr[largest])
		{
			largest = right;	
		} 
		if (largest != i)
		{
			long temp = arr[i];
			arr[i] = arr[largest];
			arr[largest] = temp;	
			heapify (arr, largest, heapSize);	
		}
	}
	
	public void sort (short[] arr)
	{
		sort (arr, 0, arr.length-1);
	}
	
	public void sort (short[] arr, int begin, int end)
	{
		buildHeap (arr, begin, end);
		short temp;
		for (int i = (end-begin); i > begin; i--)
		{
			temp = arr[begin];
			arr[begin] = arr[i];
			arr[i] = temp;
			heapify (arr, begin, i-1);
		}	
	}
	
	protected void buildHeap (short[] arr, int begin, int end)
	{
		int n = end-begin;
		for (int i = (n / 2); i >= begin; i--)
		{
			heapify (arr, i, n);	
		}
	}
	
	protected void heapify (short[] arr, int i, int heapSize)
	{
		int left = 2*i;
		int right = (2*i)+1;
		int largest;
		
		if (left <= heapSize && arr[left] > arr[i])
		{
			largest = left;
		}
		else
		{
			largest = i;	
		}
		if (right <= heapSize && arr[right] > arr[largest])
		{
			largest = right;	
		} 
		if (largest != i)
		{
			short temp = arr[i];
			arr[i] = arr[largest];
			arr[largest] = temp;	
			heapify (arr, largest, heapSize);	
		}
	}

	public void sort (double[] arr)
	{
		sort (arr, 0, arr.length-1);
	}
	
	public void sort (double[] arr, int begin, int end)
	{
		buildHeap (arr, begin, end);
		double temp;
		for (int i = (end-begin); i > begin; i--)
		{
			temp = arr[begin];
			arr[begin] = arr[i];
			arr[i] = temp;
			heapify (arr, begin, i-1);
		}	
	}
	
	protected void buildHeap (double[] arr, int begin, int end)
	{
		int n = (end-begin);
		for (int i = (n / 2); i >= begin; i--)
		{
			heapify (arr, i, n);	
		}
	}
	
	protected void heapify (double[] arr, int i, int heapSize)
	{
		int left = 2*i;
		int right = (2*i)+1;
		int largest;
		
		if (left <= heapSize && arr[left] > arr[i])
		{
			largest = left;
		}
		else
		{
			largest = i;	
		}
		if (right <= heapSize && arr[right] > arr[largest])
		{
			largest = right;	
		} 
		if (largest != i)
		{
			double temp = arr[i];
			arr[i] = arr[largest];
			arr[largest] = temp;	
			heapify (arr, largest, heapSize);	
		}
	}

	public void sort (float[] arr)
	{
		sort (arr, 0, arr.length-1);
	}
	
	public void sort (float[] arr, int begin, int end)
	{
		buildHeap (arr, begin, end);
		float temp;
		for (int i = (end-begin); i > begin; i--)
		{
			temp = arr[begin];
			arr[begin] = arr[i];
			arr[i] = temp;
			heapify (arr, begin, i-1);
		}	
	}
	
	protected void buildHeap (float[] arr, int begin, int end)
	{
		int n = (end - begin);
		for (int i = (n / 2); i >= begin; i--)
		{
			heapify (arr, i, n);	
		}
	}
	
	protected void heapify (float[] arr, int i, int heapSize)
	{
		int left = 2*i;
		int right = (2*i)+1;
		int largest;
		
		if (left <= heapSize && arr[left] > arr[i])
		{
			largest = left;
		}
		else
		{
			largest = i;	
		}
		if (right <= heapSize && arr[right] > arr[largest])
		{
			largest = right;	
		} 
		if (largest != i)
		{
			float temp = arr[i];
			arr[i] = arr[largest];
			arr[largest] = temp;	
			heapify (arr, largest, heapSize);	
		}
	}

	public void sort (char[] arr)
	{
		sort (arr, 0, arr.length-1);
	}
	
	public void sort (char[] arr, int begin, int end)
	{
		buildHeap (arr, begin, end);
		char temp;
		for (int i = (end - begin); i > begin; i--)
		{
			temp = arr[begin];
			arr[begin] = arr[i];
			arr[i] = temp;
			heapify (arr, begin, i-1);
		}	
	}
	
	protected void buildHeap (char[] arr, int begin, int end)
	{
		int n = (end - begin);
		for (int i = (n / 2); i >= begin; i--)
		{
			heapify (arr, i, n);	
		}
	}
	
	protected void heapify (char[] arr, int i, int heapSize)
	{
		int left = 2*i;
		int right = (2*i)+1;
		int largest;
		
		if (left <= heapSize && arr[left] > arr[i])
		{
			largest = left;
		}
		else
		{
			largest = i;	
		}
		if (right <= heapSize && arr[right] > arr[largest])
		{
			largest = right;	
		} 
		if (largest != i)
		{
			char temp = arr[i];
			arr[i] = arr[largest];
			arr[largest] = temp;	
			heapify (arr, largest, heapSize);	
		}
	}

	public void sort (byte[] arr)
	{
		sort (arr, 0, arr.length-1);
	}
	
	public void sort (byte[] arr, int begin, int end)
	{
		buildHeap (arr, begin, end);
		byte temp;
		for (int i = (end - begin); i > begin; i--)
		{
			temp = arr[begin];
			arr[begin] = arr[i];
			arr[i] = temp;
			heapify (arr, begin, i-1);
		}	
	}
	
	protected void buildHeap (byte[] arr, int begin, int end)
	{
		int n = (end - begin);
		for (int i = (n / 2); i >= begin; i--)
		{
			heapify (arr, i, n);	
		}
	}
	
	protected void heapify (byte[] arr, int i, int heapSize)
	{
		int left = 2*i;
		int right = (2*i)+1;
		int largest;
		
		if (left <= heapSize && arr[left] > arr[i])
		{
			largest = left;
		}
		else
		{
			largest = i;	
		}
		if (right <= heapSize && arr[right] > arr[largest])
		{
			largest = right;	
		} 
		if (largest != i)
		{
			byte temp = arr[i];
			arr[i] = arr[largest];
			arr[largest] = temp;	
			heapify (arr, largest, heapSize);	
		}
	}

	public void sort (Object[] arr, Comparator c, String methodName) throws SortException
	{
		sort (arr, 0, arr.length-1, c, methodName, new Class[0], new Object[0]);	
	}
	
	public void sort (Object[] arr, int begin, int end, Comparator c, String methodName) throws SortException
	{
		sort (arr, begin, end, c, methodName, new Class[0], new Object[0]);	
	}
	
	public void sort (Object[] arr, Comparator c, String methodName, Class[] paramTypes, Object[] paramValues)
						 throws SortException
	{
		sort (arr, 0, arr.length-1, c, methodName, paramTypes, paramValues);	
	}
	
	public void sort (Object[] arr, int begin, int end, Comparator c, String methodName, Class[] paramTypes, Object[] paramValues)
						 throws SortException
	{
		try
		{
			buildHeap (arr, begin, end, c, methodName, paramTypes, paramValues);
			Object temp;
			for (int i = (end-begin); i > begin; i--)
			{
				temp = arr[begin];
				arr[begin] = arr[i];
				arr[i] = temp;
				heapify (arr, begin, i-1, c, methodName, paramTypes, paramValues);
			}	
		}
		catch (Exception e)
		{
			throw new SortException (e);	
		}
	}
	
	protected void buildHeap (Object[] arr, int begin, int end, Comparator c, String methodName, Class[] paramTypes, Object[] paramValues)
		throws Exception
	{
		int n = (end-begin);
		for (int i = (n / 2); i >= begin; i--)
		{
			heapify (arr, i, n, c, methodName, paramTypes, paramValues);	
		}
	}
	
	protected void heapify (Object[] arr, int i, int heapSize, Comparator c, String methodName, Class[] paramTypes, 
		Object[] paramValues) throws Exception
	{
		int left = 2*i;
		int right = (2*i)+1;
		int largest;
		
		if (left <= heapSize && c.compare(ReflectionUtil.getMethodValue(arr[left], methodName, paramTypes, paramValues),
			ReflectionUtil.getMethodValue(arr[i], methodName, paramTypes, paramValues)) > 0)
		{
			largest = left;
		}
		else
		{
			largest = i;	
		}
		if (right <= heapSize && c.compare(ReflectionUtil.getMethodValue(arr[right], methodName, paramTypes, paramValues),
			ReflectionUtil.getMethodValue(arr[largest], methodName, paramTypes, paramValues)) > 0)
		{
			largest = right;	
		} 
		if (largest != i)
		{
			Object temp = arr[i];
			arr[i] = arr[largest];
			arr[largest] = temp;	
			heapify (arr, largest, heapSize, c, methodName, paramTypes, paramValues);	
		}
	}

	public void sort (Object[] arr, Comparator c)
	{
		sort(arr, 0, arr.length-1, c);
	}
	
	public void sort (Object[] arr, int begin, int end, Comparator c)
	{
		buildHeap (arr, begin, end, c);
		Object temp;
		for (int i = (end-begin); i > begin; i--)
		{
			temp = arr[begin];
			arr[begin] = arr[i];
			arr[i] = temp;
			heapify (arr, begin, i-1, c);
		}	
	}
	
	protected void buildHeap (Object[] arr, int begin, int end, Comparator c)
	{
		int n = (end-begin);
		for (int i = (n / 2); i >= begin; i--)
		{
			heapify (arr, i, n, c);	
		}
	}
	
	protected void heapify (Object[] arr, int i, int heapSize, Comparator c)
	{
		int left = 2*i;
		int right = (2*i)+1;
		int largest;
		
		if (left <= heapSize && c.compare(arr[left], arr[i]) > 0)
		{
			largest = left;
		}
		else
		{
			largest = i;	
		}
		if (right <= heapSize && c.compare(arr[right], arr[largest]) > 0)
		{
			largest = right;	
		} 
		if (largest != i)
		{
			Object temp = arr[i];
			arr[i] = arr[largest];
			arr[largest] = temp;	
			heapify (arr, largest, heapSize, c);	
		}
	}
	
	public void sort (List arr, Comparator c)
	{
		sort(arr, 0, arr.size()-1, c);
	}
	
	public void sort (List arr, int begin, int end, Comparator c)
	{
		buildHeap (arr, begin, end, c);
		Object temp;
		for (int i = (end-begin); i > begin; i--)
		{
			temp = arr.get(begin);
			arr.set(begin, arr.get(i));
			arr.set(i, temp);
			heapify (arr, begin, i-1, c);
		}	
	}
	
	protected void buildHeap (List arr, int begin, int end, Comparator c)
	{
		int n = (end-begin);
		for (int i = (n / 2); i >= begin; i--)
		{
			heapify (arr, i, n, c);	
		}
	}
	
	protected void heapify (List arr, int i, int heapSize, Comparator c)
	{
		int left = 2*i;
		int right = (2*i)+1;
		int largest;
		
		if (left <= heapSize && c.compare(arr.get(left), arr.get(i)) > 0)
		{
			largest = left;
		}
		else
		{
			largest = i;	
		}
		if (right <= heapSize && c.compare(arr.get(right), arr.get(largest)) > 0)
		{
			largest = right;	
		} 
		if (largest != i)
		{
			Object temp = arr.get(i);
			arr.set(i, arr.get(largest));
			arr.set(largest, temp);	
			heapify (arr, largest, heapSize, c);	
		}
	}
}